"""
The module contains the BatchingProcessor class, which performs the work of grouping tasks for processing in one batch.

To use it, you need to define a class/function that will be directly called from the batcher to execute a batch of data.

Example 1:
----------
class Estimation:
    def __init__(self, executor):
        self.executor = executor

    async def estimate(self, data: list):
        await self.executor(data)

estimation = Estimation(...)
batcher = BatchingProcessor(estimation.estimate, ...)
batcher.start()

Example 2:
----------
async def estimate(data: list):
    ...

batcher = BatchingProcessor(estimate, ...)
batcher.start()

"""

import asyncio
from dataclasses import dataclass, field
from typing import Any, Awaitable, Callable


@dataclass(unsafe_hash=True)
class BatchJob:
    """
    Batch data Job.

    Attributes:
        data: any batch data
        future (asyncio.Future): futures for tracking completion of work
    """

    data: Any
    future: asyncio.Future = field(init=False)

    def __post_init__(self):
        self.future = asyncio.Future()

    async def getResult(self) -> Any:
        """Get job result."""
        await self.future
        return self.future.result()

    def setResult(self, data) -> None:
        """Set job result."""
        self.future.set_result(data)

    def setException(self, e) -> None:
        """Set job exception."""
        self.future.set_exception(e)

    def cancel(self) -> None:
        """Cancel job future."""
        self.future.cancel()

    def done(self) -> bool:
        """Check job future for completion."""
        return self.future.done()

    def cancelled(self) -> bool:
        """Check job future cancellation."""
        return self.future.cancelled()


class FlexibleBatchQueue(asyncio.Queue):
    """
    Queue generating batches of non-constant size (batchSize >= len(batch) >= 1)

    Attributes:
        batchSize (int): maximum batch size
        _eventNotEmpty (asyncio.Event): event not empty queue
        _waitEvents (set[asyncio.Event]): set of events for receiving a batch. Provides order of access to batches
    """

    def __init__(self, batchSize: int, maxQueueSize: int = 0):
        super().__init__(maxQueueSize)
        self.batchSize: int = batchSize
        self._lock: asyncio.Lock = asyncio.Lock()

    async def get(self) -> list:
        """Get batch. Wait if queue empty."""
        async with self._lock:
            if self.qsize() > 0:
                return [self.get_nowait() for _ in range(min(self.batchSize, self.qsize()))]
            return [await super().get()]


@dataclass
class BatchWorker[T, P]:
    """
    Batch Worker.

    Attributes:
        queue (FlexibleBatchQueue): queue with batches
        processor (Callable[[list], Awaitable[list]]): the called object that directly performs work on the batch
        _isRunning (bool): is BatchWorker working
        _runningTask (asyncio.Task): main loop task
        _currentBatchTask (asyncio.Task): current batching task
    """

    queue: FlexibleBatchQueue
    processor: Callable[[list | P], Awaitable[list | T]]
    _isRunning: bool = field(init=False, default=False)
    _runningTask: asyncio.Task | None = field(init=False, default=None)
    _currentBatchTask: asyncio.Task | None = field(init=False, default=None)

    async def _runBatch(self, batch: list[BatchJob]) -> None:
        """
        Batch runner.

        Args:
            batch (list[BatchJob]): batch for process
        """
        try:
            for job, result in zip(batch, await self.processor([job.data for job in batch])):
                if not job.done():  # avoid errors if task was completed/cancelled
                    job.setResult(result)
        except Exception as e:
            for job in batch:
                if not job.done():
                    job.setException(e)

    async def _runner(self) -> None:
        """Main loop task."""
        while self._isRunning:
            batch = await self.queue.get()
            await self._runBatch(batch)

    def start(self) -> None:
        """Start BatchWorker"""
        self._isRunning = True
        self._runnerTask = asyncio.create_task(self._runner())

    def terminate(self) -> None:
        """Stop BatchWorker. The current batching task will also be stopped."""
        self._isRunning = False
        if self._runnerTask:
            self._runnerTask.cancel()


@dataclass
class BatchingProcessor[T, P]:
    """
    Batching Processor.

    Attributes:
        processor (Callable[[list], Awaitable[list]]): the called object that directly performs work on the batch
        batchSize (int): maximum batch size
        numWorkers (int): workers count
        _queue (FlexibleBatchQueue): batch generator
        _workers (list[BatchWorker]): list workers
        _jobs (set[BatchJob]): set of running jobs
        _isRunning (bool): is BatchingProcessor working
    """

    processor: Callable[[list | P], Awaitable[list | T]]
    batchSize: int
    numWorkers: int

    _queue: FlexibleBatchQueue = field(init=False)
    _workers: list[BatchWorker] = field(init=False, default_factory=lambda: [])
    _jobs: set[BatchJob] = field(init=False, default_factory=lambda: set())
    _isRunning: bool = field(init=False, default=False)

    def __post_init__(self):
        if not self.batchSize > 0:
            raise ValueError("batchSize must be >= 1")
        if not self.numWorkers > 0:
            raise ValueError("numWorkers must be >= 1")
        self._queue = FlexibleBatchQueue(self.batchSize)

    async def _newJob(self, payload: P) -> T:
        """
        Add new job for batching.

        Args:
            payload: any data for batching
        """
        job = BatchJob(payload)
        await self._queue.put(job)
        self._jobs.add(job)
        try:
            result = await job.getResult()
        finally:
            self._jobs.discard(job)
        return result

    def start(self) -> None:
        """Start batching."""
        if self._isRunning:
            return
        self._isRunning = True
        self._workers = [BatchWorker(self._queue, self.processor) for _ in range(self.numWorkers)]
        for worker in self._workers:
            worker.start()

    def terminate(self) -> None:
        """Stop BatchingProcessor. Stop all tasks."""
        self._isRunning = False
        for worker in self._workers:
            worker.terminate()
        for job in self._jobs:
            job.cancel()
        self._jobs = set()
        self._workers = []

    async def put(self, payload: P) -> T:
        """
        Put data for batching and wait result.

        Args:
            payload: any data for batching.

        Raises:
            RuntimeError: if the BatchingProcessor was not started/suspended

        Returns:
            batching result

        """
        if not self._isRunning:
            raise RuntimeError("Batching stopped, adding new tasks is not allowed")
        return await self._newJob(payload)
