from typing import Generic, TypeVar
from uuid import uuid4

import attr
from luna_analytics_manager.templates.models import EventStatus, TrackStatus, VideoSegment
from lunavl.sdk.image_utils.image import VLImage

from .protocols import Event as EventProtocol

EventT = TypeVar("EventT", bound=EventProtocol)


@attr.dataclass(slots=True)
class Overview:
    """Snapshot extended with info about frame."""

    #  frame time
    timeOffset: float
    # origin image
    origin: VLImage
    # image
    snapshot: bytes | None = None

    def __deepcopy__(self, memo):
        return self.__class__(timeOffset=self.timeOffset, origin=self.origin, snapshot=self.snapshot)

    def __bool__(self):
        return self.snapshot is not None

    def asDict(self):
        """Build results in API format."""

        res = {
            "time_offset": self.timeOffset,
            "image": self.snapshot,
        }
        return res


OverviewT = TypeVar("OverviewT", bound=Overview)


@attr.dataclass(slots=True)
class AggregatedEvent(Generic[OverviewT]):
    """Container for event aggregated result."""

    # event type
    eventType: str
    #  event overview
    overview: OverviewT
    # track id
    trackId: str | None = None
    # event id
    eventId: str | None = None
    # start and end of event
    videoSegment: VideoSegment | None = None

    def asDict(self):
        """Build results in API format."""

        result = {
            "track_id": self.trackId,
            "event_id": self.eventId,
            "video_segment": self.videoSegment.asDict(),
        }
        if self.overview:
            result["overview"] = self.overview.asDict()
        return result


AggregatedEventT = TypeVar("AggregatedEventT", bound=AggregatedEvent)


@attr.dataclass(slots=True)
class Event(Generic[AggregatedEventT]):
    """Event container in processing. Event state after another frame processing"""

    eventType: str
    # aggregated event
    aggregated: AggregatedEventT | None = None
    # track id
    trackId: str | None = None
    # event id
    eventId: str | None = attr.field(factory=lambda: str(uuid4()))
    # current track status
    trackStatus: TrackStatus = TrackStatus.started
    # current event status
    eventStatus: EventStatus = EventStatus.started
    # start event time
    timeOffset: float | None = None

    def asDict(self):
        """Build event data in API format."""
        res = {
            "track_id": self.trackId,
            "event_id": self.eventId,
            "track_status": self.trackStatus.value,
            "event_status": self.eventStatus.value,
            "time_offset": self.timeOffset,
        }
        if self.aggregated is not None:
            res["aggregated"] = self.aggregated.asDict()
        return res


class Track(Generic[EventT, AggregatedEventT]):
    """Analytics track."""

    __slots__ = ("trackId", "lastEventTime", "currentEvent", "currentAggregatedEvent")

    def __init__(self, event: EventT, aggregatedEvent: AggregatedEventT):
        self.currentEvent: EventT = event
        self.currentAggregatedEvent: AggregatedEventT = aggregatedEvent
        self.trackId = self.currentEvent.trackId = self.currentAggregatedEvent.trackId = str(uuid4())
        self.lastEventTime: float | None = None
