from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Generic, Type, TypeVar, cast

from luna_analytics_manager.sync import Syncer2
from luna_analytics_manager.templates.mixins.frame_frequency_mixins import FramesFrequencyMixin
from luna_analytics_manager.templates.models import EventStatus, TrackStatus

from .protocols import AggregatedEvent, AnalyticsCtx, Event, FrameCtx, FrameRes, FrameTrack

FrameCtxT = TypeVar("FrameCtxT", bound=FrameCtx)
AnalyticsCtxT = TypeVar("AnalyticsCtxT", bound=AnalyticsCtx)
EventT = TypeVar("EventT", bound=Event)
AggregatedEventT = TypeVar("AggregatedEventT", bound=AggregatedEvent)
FrameResT = TypeVar("FrameResT", bound=FrameRes)
Estimation = TypeVar("Estimation")
TrackT = TypeVar("TrackT", bound=FrameTrack, covariant=True)


class BaseFrameAnalyticsNode(
    Generic[FrameCtxT, AnalyticsCtxT, EventT, AggregatedEventT, FrameResT, Estimation, TrackT],
    FramesFrequencyMixin,
    ABC,
):
    """Base class for simple analytics node based on `frame estimation`.
    It is assumed that the analytics is estimating or detecting something in the entire frame or some area of it.

    Example: crowd, fire, ... We want to determinate there is a fire on video or not.

    The analytics has no more than one track every moment. The track begins when an interesting object is first
    discovered (+probe count) and ends when the object of interest is no longer observed.

    Supports several event generation policy out the box:

        1. start - first an object discovering.
        2. end - when the object of interest is no longer observed.
        3. period - periodical event generation.

    Attributes:
        track: current track.
        trackLength: current track length.
        frameSyncer: syncer
    """

    # class for result by frame
    frameResCls: Type[FrameResT]
    # class for tracks
    trackCls: Type[TrackT]

    def __init__(self):
        super().__init__()
        self.frameSyncer = Syncer2()
        self.track: TrackT | None = None
        self.trackLength = 0

    @abstractmethod
    async def estimate(self, frameCtx: FrameCtxT, ctx: AnalyticsCtxT):
        """Estimate something and apply after estimation logic (DROI, rescale, ...)."""
        ...

    @abstractmethod
    def updateAggregatedEvent(self, estimation: Estimation, frameCtx: FrameCtxT, ctx: AnalyticsCtxT):
        """
        Update aggregated event from estimation
        """
        ...

    @abstractmethod
    def buildEvent(
        self, estimation: Estimation, frameCtx: FrameCtxT, ctx: AnalyticsCtxT
    ) -> tuple[EventT, AggregatedEventT]:
        """
        Generate new event and aggregated event from estimation.

        Call when current track has no current event.
        """
        ...

    @abstractmethod
    def isTrackNeedEnd(self, estimation: Estimation, frameCtx: FrameCtxT, ctx: AnalyticsCtxT) -> bool:
        """
        Determinate that event is needed to end or not after current frame processing.

        If function return `True`, track will be stopped. If function return `False` track will be started or continue.
        """
        ...

    def updateTrack(self, estimation: Estimation, frameCtx: FrameCtxT, ctx: AnalyticsCtxT):
        """Update current track with a new frame result."""
        track = self.track
        if not track:
            self.track = self.trackCls(*self.buildEvent(estimation, frameCtx, ctx))
        elif not track.currentEvent:
            event, aggregatedEvent = self.buildEvent(estimation, frameCtx, ctx)
            event.trackId = aggregatedEvent.trackId = track.trackId
            event.trackStatus = TrackStatus.inProcess
            track.currentEvent = event
            track.currentAggregatedEvent = aggregatedEvent
        else:
            self.updateAggregatedEvent(estimation, frameCtx, ctx)
            self.updateEvent(estimation, frameCtx, ctx)

    def updateEvent(self, estimation: Estimation, frameCtx: FrameCtxT, ctx: AnalyticsCtxT):
        """Update current event with new frame."""
        track = cast(TrackT, self.track)
        event = track.currentEvent
        event.estimation = estimation
        event.eventStatus = EventStatus.inProcess

    def closeTrack(self):
        """Close current track, remove data"""

        self.trackLength = 0
        self.track = None

    async def run(self, frameCtx: FrameCtxT, ctx: AnalyticsCtxT):
        """Run template node processing."""

        async with self.frameSyncer.process(frameCtx.number) as lock:
            if self.checkFrameSkip(frameCtx, ctx):
                if frameCtx.lastFrame:
                    async with lock.acquire():
                        if self.track and self.track.currentEvent:
                            self.track.currentEvent.trackStatus = TrackStatus.finished
                            currentEvent = await self.processEvent(frameCtx, ctx)
                            frameCtx.processed = self.frameResCls(events=[currentEvent])
                        self.closeTrack()
                return

            template = await self.estimate(frameCtx, ctx)
            if template is None:
                return

            async with lock.acquire():
                await self.postProcessing(template, frameCtx, ctx)

    async def postProcessing(self, estimation: Estimation, frameCtx: FrameCtxT, ctx: AnalyticsCtxT):
        """Post processing."""

        if self.isTrackNeedEnd(estimation, frameCtx, ctx):
            if self.track and self.track.currentEvent:
                # there is active track needed to be finalized
                self.track.currentEvent.trackStatus = TrackStatus.finished
            else:
                self.closeTrack()
                return
        elif self.trackLength < ctx.params.probeCount - 1:
            # wait probe count
            self.trackLength += 1
            return
        else:
            self.updateTrack(estimation, frameCtx, ctx)
        track = cast(TrackT, self.track)
        if frameCtx.lastFrame:
            track.currentEvent.trackStatus = TrackStatus.finished

        currentEvent = await self.processEvent(frameCtx, ctx)
        if not frameCtx.processed:
            frameCtx.processed = self.frameResCls(events=[currentEvent])
        else:
            frameCtx.processed.events.append(currentEvent)

        if track and (self.isTrackNeedEnd(estimation, frameCtx, ctx) or frameCtx.lastFrame):
            self.closeTrack()

    async def finalizeEvent(self, frameCtx: FrameCtxT) -> Event:
        """Finalize event."""
        track = cast(TrackT, self.track)
        event = track.currentEvent
        event.eventStatus = EventStatus.finished
        event.aggregated = track.currentAggregatedEvent
        track.currentEvent = None
        track.currentAggregatedEvent = None
        track.lastEventTime = frameCtx.timestamp
        return event

    async def processEvent(self, frameCtx: FrameCtxT, ctx: AnalyticsCtxT) -> Event:
        """Process current event."""
        track = cast(TrackT, self.track)
        event = track.currentEvent
        eventPolicy = ctx.params.eventPolicy
        if eventPolicy.trigger == "start" and event.trackStatus != TrackStatus.started:
            pass
        elif eventPolicy.trigger == "end" and event.trackStatus != TrackStatus.finished:
            pass
        elif (
            eventPolicy.trigger == "period"
            and track.lastEventTime is not None
            and (frameCtx.timestamp - track.lastEventTime) < eventPolicy.interval
        ):
            pass
        else:
            event.eventStatus = EventStatus.finished

        if event.eventStatus == EventStatus.finished:
            currentEvent = await self.finalizeEvent(frameCtx)
        else:
            currentEvent = track.currentEvent

        # copy data, self.track.currentEvent may be change in next frame
        return deepcopy(currentEvent)
