from abc import ABC, abstractmethod
from typing import Annotated, Literal

from aiohttp import BasicAuth
from annotated_types import Ge
from pydantic import Field, Strict
from vlutils.structures.pydantic import BaseModel

from .luna_wraps import LunaContext


class BasicAuthorization(BaseModel):
    """
    Callback basic authorization

    >>> print(BasicAuthorization(**{"type": "basic", "login": "login", "password": "password"}).headers)
    {'Authorization': 'Basic bG9naW46cGFzc3dvcmQ='}
    """

    type: Literal["basic"]
    login: Annotated[str, Field(max_length=128, min_length=3)]
    password: Annotated[str, Field(max_length=128, min_length=3)]

    @property
    def headers(self):
        """Get basic authorization headers"""
        return {"Authorization": BasicAuth(login=self.login, password=self.password, encoding="utf-8").encode()}


class HTTPCallbackParams(BaseModel):
    """HTTP Callback request parameters"""

    timeout: Annotated[int, Ge(0), Strict()] = Field(default_factory=lambda: None)
    contentType: Literal["application/json"] = Field(default_factory=lambda: None)
    headers: dict[str, str] = Field(default_factory=lambda: None)


class Trigger(BaseModel):
    """
    Callback trigger
    """

    def fitCondition(self, enrichedEvent: dict) -> bool:
        """Whether event fit trigger condition(s)"""
        return True


class BaseCallback(BaseModel, ABC):
    """
    Base callback
    """

    type: str
    enable: Literal[0, 1] = 1
    trigger: Trigger = Trigger()

    def isNeedToExecute(self, enrichedEvent: dict) -> bool:
        """Whether callback must be executed"""
        if not self.enable:
            return False
        if hasattr(self, "trigger") and hasattr(self.trigger, "fitCondition"):
            return self.trigger.fitCondition(enrichedEvent=enrichedEvent)
        return True

    @abstractmethod
    def getCredentialsKey(self) -> str:
        """Get data which is a request characterization (what and where to send) | defines uniqueness of callbacks"""

    @abstractmethod
    async def execute(self, lunaContext: LunaContext, enrichedEvent: dict) -> None:
        """Callback execution coro"""


class HttpAnalyticCallback(BaseCallback):
    """Http analytics callback"""

    url: Annotated[str, Field(min_length=3)]
    type: Literal["http"]
    authorization: BasicAuthorization = Field(default_factory=lambda: None)
    params: HTTPCallbackParams = HTTPCallbackParams()

    def getCredentialsKey(self) -> str:
        """See description of `BaseCallback.getCredentialsKey`"""
        if getattr(self, "_credentialsKey", None) is None:
            self._credentialsKey = f"{self.url},{self.authorization},{self.params.headers},{self.params.contentType}"
        return self._credentialsKey

    async def execute(self, lunaContext: LunaContext, enrichedEvent: dict) -> None:
        """See description of `BaseCallback.execute`"""
        headers = {"Content-Type": "application/json", "Luna-Request-Id": lunaContext.requestId}
        if self.params.headers is not None:
            headers |= self.params.headers
        if self.params.contentType is not None:
            headers |= {"Content-Type": self.params.contentType}
        if self.authorization is not None:
            headers |= self.authorization.headers
        await lunaContext.httpClient.makeRequest(
            url=self.url,
            method="POST",
            body=enrichedEvent,
            headers=headers,
            totalTimeout=self.params.timeout,
            asyncRequest=True,
            raiseError=True,
        )


class WSAnalyticCallback(BaseCallback):
    """
    Websocket analytics callback
    """

    type: Literal["luna-ws-notification"]

    def getCredentialsKey(self) -> str:
        """See description of `BaseCallback.getCredentialsKey`"""
        if getattr(self, "_credentialsKey", None) is None:
            self._credentialsKey = str(id(self))
        return self._credentialsKey

    async def execute(self, lunaContext: LunaContext, enrichedEvent: dict) -> None:
        """See description of `BaseCallback.execute`"""
        if not lunaContext.senderClient:
            return
        await lunaContext.senderClient.generalPublishOne(
            event=enrichedEvent,
            lunaRequestId=lunaContext.requestId,
            asyncRequest=True,
            raiseError=True,
        )


class EventAnalyticCallback(BaseCallback):
    """
    Event analytics callback
    """

    type: Literal["luna-event"]

    def getCredentialsKey(self) -> str:
        """See description of `BaseCallback.getCredentialsKey`"""
        if getattr(self, "_credentialsKey", None) is None:
            self._credentialsKey = str(id(self))
        return self._credentialsKey

    async def execute(self, lunaContext: LunaContext, enrichedEvent: dict) -> None:
        """See description of `BaseCallback.execute`"""
        if not lunaContext.eventsClient:
            return
        await lunaContext.eventsClient.saveGeneralEvents(
            events=[enrichedEvent],
            waitEventsSaving=False,
            lunaRequestId=lunaContext.requestId,
            asyncRequest=True,
            raiseError=True,
        )


DefaultCallbacks = Annotated[
    HttpAnalyticCallback | WSAnalyticCallback | EventAnalyticCallback, Field(discriminator="type")
]
