from typing import Generic, TypeVar

from luna_analytics_manager.templates.image_helpers import toBytes
from luna_analytics_manager.templates.image_helpers.roi import getROIRect, scaleDROI
from luna_analytics_manager.templates.mixins.frame_frequency_mixins import FramesFrequencyMixin
from luna_analytics_manager.templates.sdk import SDKAnalyticsCtx, SDKFrameCtx
from lunavl.sdk.detectors.base import ImageForDetection
from lunavl.sdk.detectors.bodydetector import BodyDetection
from shapely import Point

from .protocols import AnalyticsCtx, FrameCtx

FrameCtxT = TypeVar("FrameCtxT", bound=FrameCtx)
AnalyticsCtxT = TypeVar("AnalyticsCtxT", bound=AnalyticsCtx)


class NodeOverview(Generic[FrameCtxT, AnalyticsCtxT]):
    """Base overview node."""

    name = "overview"

    async def run(self, frameCtx: FrameCtxT, ctx: AnalyticsCtxT):
        """Run overview node processing."""

        if not frameCtx.processed:
            return

        if not (event := frameCtx.processed.events[0]).aggregated:
            return

        overview = event.aggregated.overview
        overview.snapshot, _ = await toBytes(event.aggregated.overview.origin, ctx.params.imageRetainPolicy)


class NodeFrameSkipper(Generic[FrameCtxT, AnalyticsCtxT], FramesFrequencyMixin):
    """Helpers node for global"""

    name = "frame_skipper"

    async def run(self, frameCtx: FrameCtx, ctx: AnalyticsCtx):
        frameCtx.skip = self.checkFrameSkip(frameCtx, ctx)


class _SkipperMeta(type):
    """
    Metaclass patching node class for skipping frame process
    """

    def __call__(cls):
        classInstance = super().__call__()

        oldRun = classInstance.run

        async def newRun(frameCtx, *args, **kwargs):
            if not frameCtx.skip:
                return await oldRun(frameCtx, *args, **kwargs)
            return

        classInstance.run = newRun
        return classInstance


class SkipperMixin(metaclass=_SkipperMeta):
    """
    Mixin for skipping frame if it has skipp attribute with value True

    >>> import asyncio
    >>> class FooNode(SkipperMixin):
    >>>     requires = []
    >>>
    >>>     async def run(self, frameCtx, ctx):
    >>>         print("run")
    >>>
    >>> class Frame:
    >>>     def __init__(self, skip: bool, data: None | dict = None):
    >>>         self.skip = skip
    >>> node = FooNode()
    >>> frameCtx = Frame(skip=True)
    >>> asyncio.run(node.run(frameCtx, None))
    "run"
    >>> frameCtx = Frame(skip=False)
    >>> node.run(frameCtx, None)

    """


def centerPoint(rect) -> Point:
    """Get cent point of point"""
    return Point(rect.x + rect.width / 2, rect.y + rect.height / 2)


class NodeBodyDetection:
    """
    Node for body detection.

    The Node detects body, applies roi and droi parameters and set them to  frame ctx as list[BodyDetection]
    """

    name = "body_detection"

    def getBodyPoint(self, detection: BodyDetection):
        """
        Get body point  by detection for  further filtration.

        Function returns a center of detection now,  user can overload this function and use lower left corner
        of detection.
        """
        return centerPoint(detection.boundingBox.rect)

    async def run(self, frameCtx: SDKFrameCtx, ctx: SDKAnalyticsCtx):

        detector = ctx.sdk.engine.estimators.getBodyDetector()

        image = frameCtx.image
        if ctx.params.roi:
            rect = getROIRect(frameCtx.image.rect, ctx.params.roi)
            image = ImageForDetection(image, rect)

        frameCtx.bodyDetections = await detector.estimate(image)

        if ctx.params.droi:
            droi = scaleDROI(frameCtx.image.rect, ctx.params.droi)
            frameCtx.bodyDetections = list(
                filter(lambda body: droi.covers(self.getBodyPoint(body)), frameCtx.bodyDetections)
            )
