from numbers import Real
from typing import Awaitable, Protocol, TypeVar

from luna3.common.luna_response import LunaResponse
from luna_analytics_manager.base import AnalyticsCtx as BaseAnalyticsCtx, FrameCtx as FrameCtxBase
from luna_analytics_manager.templates.models import EventStatus, TrackStatus, VideoSegment
from lunavl.sdk.image_utils.image import VLImage

from .params import BaseParams


class Overview(Protocol):
    #  frame time
    timeOffset: float
    # origin image
    origin: VLImage
    # image
    snapshot: bytes | None = None

    def asDict(self) -> dict:
        pass

    def __bool__(self):
        """Fill overview or not"""


class AggregatedEvent(Protocol):
    """Aggregated event protocol"""

    # event type
    eventType: str
    # track id
    trackId: str | None
    # event id
    eventId: str | None
    # video segment, start and end of event
    videoSegment: VideoSegment | None
    # overview shot
    overview: Overview

    def asDict(self) -> dict:
        pass


AggregatedEventT = TypeVar("AggregatedEventT", bound=AggregatedEvent)


class Event(Protocol[AggregatedEventT]):
    """Protocol for current event"""

    # event type
    eventType: str
    # track id
    trackId: str | None
    # event id
    eventId: str | None
    # current track status
    trackStatus: TrackStatus
    # current event status
    eventStatus: EventStatus
    # accumulating aggregated event from this event
    aggregated: AggregatedEventT | None
    # current event time
    timeOffset: float | None

    def asDict(self) -> dict:
        pass


EventT = TypeVar("EventT", bound=Event)


class FrameRes(Protocol[EventT]):
    """Frame result protocol"""

    # list of current events
    events: list[EventT]

    def asDict(self) -> dict:
        pass


class FrameTrack(Protocol[EventT, AggregatedEventT]):
    """Protocol for track based frame estimation"""

    # current track event
    currentEvent: EventT
    # last event time (time based on video data)
    lastEventTime: float | None
    # current aggregated event
    currentAggregatedEvent: AggregatedEventT
    # track id
    trackId: str


FrameResT = TypeVar("FrameResT", bound=FrameRes, covariant=True)


class FrameCtx(FrameCtxBase, Protocol[FrameResT]):
    # Frame context protocol for frame base analytics
    # frame ts
    timestamp: float
    # frame number
    number: int
    # analytics result by frame
    processed: FrameResT | None
    # flag last frame or not
    lastFrame: bool


class AnalyticsCtx(BaseAnalyticsCtx):
    # analytics params
    params: BaseParams


class HTTPClient(Protocol):
    # http client protocol

    async def makeRequest(
        self,
        url: str,
        method: str,
        queryParams: dict | None = None,
        body: str | bytes | dict | list = None,
        headers: dict | None = None,
        asyncRequest: bool | None = None,
        totalTimeout: Real | None = object(),
        connectTimeout: Real | None = object(),
        sockReadTimeout: Real | None = object(),
        sockConnectTimeout: Real | None = object(),
        **kwargs,
    ) -> Awaitable[LunaResponse] | LunaResponse:
        """Make http request"""
